// Copyright 1998-2017 Epic Games, Inc. All Rights Reserved.

#include "FOHEARTAnimNode.h"
#include "Animation/AnimNodeBase.h"
#include "Animation/AnimInstanceProxy.h"
#include "FOHEART.h"

#include <EngineGlobals.h>
#include <Runtime/Engine/Classes/Engine/Engine.h>
#include <Runtime/Engine/Public/EngineUtils.h>
#include <Runtime/Engine/Classes/Engine/StaticMeshActor.h>
#include <Runtime/CoreUObject/Public/Templates/Casts.h>



// Sets default values
FFOHEARTAnimNode::FFOHEARTAnimNode()
{
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_Hips, FName("pelvis")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_Spine2, FName("spine_01")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_Spine3, FName("spine_02")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_Spine4, FName("spine_03")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_Neck, FName("neck_01")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_Head, FName("head")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightShoulder, FName("clavicle_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightArm, FName("upperarm_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightForeArm, FName("lowerarm_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightHand, FName("hand_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftShoulder, FName("clavicle_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftArm, FName("upperarm_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftForeArm, FName("lowerarm_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftHand, FName("hand_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightUpLeg, FName("thigh_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightLeg, FName("calf_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightFoot, FName("foot_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_RightToeBase, FName("ball_r")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftUpLeg, FName("thigh_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftLeg, FName("calf_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftFoot, FName("foot_l")));
	BodyBoneMap.Push(FFOHEART_C1BoneMapStruct(EFOHEART_C1BonesEnum::VE_LeftToeBase, FName("ball_l")));

	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Thumb1, FName("thumb_01_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Thumb2, FName("thumb_02_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Thumb3, FName("thumb_03_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Index1, FName("index_01_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Index2, FName("index_02_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Index3, FName("index_03_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Middle1, FName("middle_01_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Middle2, FName("middle_02_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Middle3, FName("middle_03_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Ring1, FName("ring_01_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Ring2, FName("ring_02_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Ring3, FName("ring_03_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Pinky1, FName("pinky_01_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Pinky2, FName("pinky_02_l")));
	LeftHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Pinky3, FName("pinky_03_l")));

	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Thumb1, FName("thumb_01_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Thumb2, FName("thumb_02_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Thumb3, FName("thumb_03_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Index1, FName("index_01_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Index2, FName("index_02_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Index3, FName("index_03_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Middle1, FName("middle_01_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Middle2, FName("middle_02_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Middle3, FName("middle_03_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Ring1, FName("ring_01_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Ring2, FName("ring_02_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Ring3, FName("ring_03_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Pinky1, FName("pinky_01_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Pinky2, FName("pinky_02_r")));
	RightHandBoneMap.Push(FFOHEART_C1HandBoneMapStruct(EFOHEART_C1HandBonesEnum::VS_Pinky3, FName("pinky_03_r")));
}

FFOHEARTAnimNode::~FFOHEARTAnimNode()
{

}
TArray<FTransform> inPoses;
FPoseLink InPose;
void FFOHEARTAnimNode::Initialize_AnyThread(const FAnimationInitializeContext& Context)
{
	//FAnimNode_Base::Initialize_AnyThread(Context);
	check(Context.AnimInstanceProxy != nullptr);

	//init the datastream
	InPose.Initialize(Context);


	OwningActor = Context.AnimInstanceProxy->GetSkelMeshComponent()->GetOwner();

	FBoneContainer BoneContainer = Context.AnimInstanceProxy->GetRequiredBones();

	for (int i = 0; i < BodyBoneMap.Num(); i++)
	{
		int32 findIndex = BoneContainer.GetPoseBoneIndexForBoneName(BodyBoneMap[i].TargetBoneName);
		if (INDEX_NONE == findIndex)
		{
			FString strDebug = TEXT("match bone name not found: ") + BodyBoneMap[i].TargetBoneName.ToString();
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, strDebug);
		}
		else
		{
			MapBodyBoneIndex.Add(BodyBoneMap[i].FoheartBoneName, BoneContainer.GetCompactPoseIndexFromSkeletonIndex(findIndex));
		}
	}
	for (int i = 0; i < LeftHandBoneMap.Num(); i++)
	{
		int32 findIndex = BoneContainer.GetPoseBoneIndexForBoneName(LeftHandBoneMap[i].TargetBoneName);
		if (INDEX_NONE == findIndex)
		{
			FString strDebug = TEXT("match bone name not found: ") + LeftHandBoneMap[i].TargetBoneName.ToString();
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, strDebug);
		}
		else
		{
			MapLeftHandBoneIndex.Add(LeftHandBoneMap[i].FoheartBoneName, BoneContainer.GetCompactPoseIndexFromSkeletonIndex(findIndex));
		}
	}
	for (int i = 0; i < RightHandBoneMap.Num(); i++)
	{
		int32 findIndex = BoneContainer.GetPoseBoneIndexForBoneName(RightHandBoneMap[i].TargetBoneName);
		if (INDEX_NONE == findIndex)
		{
			FString strDebug = TEXT("match bone name not found: ") + RightHandBoneMap[i].TargetBoneName.ToString();
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, strDebug);
		}
		else
		{
			MapRightHandBoneIndex.Add(RightHandBoneMap[i].FoheartBoneName, BoneContainer.GetCompactPoseIndexFromSkeletonIndex(findIndex));
		}
	}

	TActorIterator<AFOHEARTNetWorkManager> actorItr = TActorIterator<AFOHEARTNetWorkManager>(OwningActor->GetWorld());
	for (actorItr; actorItr; ++actorItr) {
		if (actorItr) {
			AFOHEARTNetWorkManager* pTmpManager = Cast<AFOHEARTNetWorkManager>(*actorItr);
			if (pTmpManager)
			{
				//if (pTmpManager->GetName().Contains("FOHEARTNetWorkManager", ESearchCase::CaseSensitive))
				{
					pNetWorkManager = pTmpManager;
					break;
				}
			}
			else {
				GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("not found"));
			}
		}
	}

	ActorInitialRotation = OwningActor->GetActorRotation();
	ActorInitialLocation = OwningActor->GetActorLocation();
}

void FFOHEARTAnimNode::Update_AnyThread(const FAnimationUpdateContext& Context)
{
	//inPoses.Empty();
	//inPoses = Context.AnimInstanceProxy->GetSkelMeshComponent()->BoneSpaceTransforms;

	//InPose.Update(Context);
	//EvaluateGraphExposedInputs.Execute(Context);
}

void FFOHEARTAnimNode::CacheBones(const FAnimationCacheBonesContext & Context)
{
	//InPose.CacheBones(Context);
}

void FFOHEARTAnimNode::Evaluate_AnyThread(FPoseContext& Output)
{
	check(Output.AnimInstanceProxy->GetSkeleton() != nullptr);
	InPose.Evaluate(Output);

	USkeletalMeshComponent* pComponent = Output.AnimInstanceProxy->GetSkelMeshComponent();

	if (pComponent)
	{
		if (pNetWorkManager && pNetWorkManager->MotionVenusConnected)
		{
			TArray<FVector>		TranslateArray;
			TArray<FRotator>	RotatorArray;
			TArray<FQuat>		LeftHandQuat;
			TArray<FQuat>		RightHandQuat;

			bool bRet = UFOHEART_C1BPLibrary::FOHEART_C1ReadArray(pNetWorkManager, ActorName, pComponent,
				TranslateArray, RotatorArray, LeftHandQuat, RightHandQuat, BodyBoneMap, LeftHandBoneMap, RightHandBoneMap);
			if (bRet)
			{
				int32 index = UFOHEART_C1BPLibrary::FOHEART_C1GetBVHBoneIndex(EFOHEART_C1BonesEnum::VE_Hips);

				for (int i = 0; i < BodyBoneMap.Num(); i++)
				{
					FCompactPoseBoneIndex* pCompactIndex = MapBodyBoneIndex.Find(BodyBoneMap[i].FoheartBoneName);
					if (pCompactIndex)
					{
						FRotator rotator = FRotator::ZeroRotator;
						if (i< RotatorArray.Num())
							rotator = RotatorArray[i];

						FQuat BVHQuat = rotator.Quaternion();
						// = CalculateQuat(rotator.Roll, rotator.Pitch, rotator.Yaw, pNetWorkManager->Skeleton.Bones[(int32)(BoneMap[i].FoheartBoneName)].RotOrder);
						//ת
						Output.Pose[*pCompactIndex].SetRotation(BVHQuat);
						//λ
						if (BodyBoneMap[i].FoheartBoneName == EFOHEART_C1BonesEnum::VE_Hips)
						{
							FVector position = FVector::ZeroVector;
							if (i<TranslateArray.Num())
								position = TranslateArray[i];

							Output.Pose[*pCompactIndex].SetTranslation(position * ScaleBasicFoheart);
							//FString strValue = "----" + position.ToString();
							//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, strValue);
						}

					}
				}
				//UE_LOG(LogInit, Log, TEXT("--------------------------"));
				for (int i = 0; i < LeftHandBoneMap.Num(); i++)
				{
					FCompactPoseBoneIndex* pCompactIndex = MapLeftHandBoneIndex.Find(LeftHandBoneMap[i].FoheartBoneName);
					if (pCompactIndex && i<LeftHandQuat.Num())
					{
						FQuat BVHQuat = LeftHandQuat[i];
						Output.Pose[*pCompactIndex].SetRotation(BVHQuat);
					}
				}
				//UE_LOG(LogInit, Log, TEXT("LeftHandBoneMap_count: %d"), LeftHandBoneMap.Num());
				//UE_LOG(LogInit, Log, TEXT("LeftHandQuat_count: %d"), LeftHandQuat.Num());

				//UE_LOG(LogInit, Log, TEXT("--------------------------"));
				for (int i = 0; i < RightHandBoneMap.Num(); i++)
				{
					FCompactPoseBoneIndex* pCompactIndex = MapRightHandBoneIndex.Find(RightHandBoneMap[i].FoheartBoneName);
					if (pCompactIndex&& i<RightHandQuat.Num())
					{
						FQuat BVHQuat = RightHandQuat[i];
						Output.Pose[*pCompactIndex].SetRotation(BVHQuat);
					}
				}

				//pComponent->UpdateBoneBodyMapping();
				//pComponent->UpdateComponentToWorld(EUpdateTransformFlags::PropagateFromParent);
			}
		}
	}
}

//void FFOHEARTAnimNode::EvaluateComponentSpace_AnyThread(FComponentSpacePoseContext & Output)
//{
//	TMap<int32, FString > bone1;
//	bone1.Add(1, "A1");
//	bone1.Add(2, "A2");
//	bone1.Add(3, "A3");
//	//bone1.Add(4, "A4");
//
//	//Output.AnimInstanceProxy;
//}


