// Copyright 1998-2017 Epic Games, Inc. All Rights Reserved.
#include "FOHEARTBPLibrary.h"
#include "FOHEART.h"
#include "FOHEARTMath.h"
#include "Runtime/Launch/Resources/Version.h"

UFOHEART_C1BPLibrary::UFOHEART_C1BPLibrary(const FObjectInitializer& ObjectInitializer)
	: Super(ObjectInitializer)
{
	// Nothing todo
}

// Init with file and read BVH reference skeleton
bool UFOHEART_C1BPLibrary::FOHEART_C1InitBVHFile(AFOHEARTController *Controller, FString BVHFileName)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	if (Controller->Skeleton.ParseBVHReferenceFile(BVHFileName) != true)
		return false;

	return true;
}

// Init and configure BVH reference skeleton
bool UFOHEART_C1BPLibrary::FOHEART_C1Init(
	AFOHEARTController *Controller, int32 BoneNr, EFOHEARTRotOrderEnum RotationOrder,
	EFOHEARTChannelNumberEnum XPos, EFOHEARTChannelNumberEnum YPos, EFOHEARTChannelNumberEnum ZPos,
	EFOHEARTChannelNumberEnum XRot, EFOHEARTChannelNumberEnum YRot, EFOHEARTChannelNumberEnum ZRot)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	Controller->Skeleton.BoneNr = BoneNr;
	Controller->Skeleton.BonesSetRotOrder((ChannelOrderEnum)RotationOrder);
	return Controller->Skeleton.BonesSetChannels((uint8)XPos, (uint8)YPos, (uint8)ZPos, (uint8)XRot, (uint8)YRot, (uint8)ZRot);
}

// Init and configure BVH reference skeleton
bool UFOHEART_C1BPLibrary::FOHEART_C1BoneSetOffset(AFOHEARTController *Controller, int32 BoneIndex, float X, float Y, float Z)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	return Controller->Skeleton.BoneSetOffset(BoneIndex, X, Y, Z);
}

// Connect to MotionVenus
bool UFOHEART_C1BPLibrary::FOHEART_C1ConnectTCP(
	AFOHEARTController *Controller, FString HostName, int32 Port,
	bool bReference, bool bDisplacement, EFOHEARTMotionLineFormatEnum MotionLineFormat
)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	Controller->bReference = bReference;
	Controller->bDisplacement = bDisplacement;
	if (MotionLineFormat == EFOHEARTMotionLineFormatEnum::VE_FOHEART_C1)
		Controller->MotionLineFormat = FOHEART_C1;
	else
		Controller->MotionLineFormat = Standard;

	if (Controller->Connect(HostName, Port) != true)
		return false;

	return true;
}

// Connect to MotionVenus
bool UFOHEART_C1BPLibrary::FOHEART_C1ConnectUDP(
	AFOHEARTController *Controller, FString HostName, int32 Port,
	bool bReference, bool bDisplacement, EFOHEARTMotionLineFormatEnum MotionLineFormat
)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	Controller->bReference = bReference;
	Controller->bDisplacement = bDisplacement;
	if (MotionLineFormat == EFOHEARTMotionLineFormatEnum::VE_FOHEART_C1)
		Controller->MotionLineFormat = FOHEART_C1;
	else
		Controller->MotionLineFormat = Standard;

	if (Controller->ConnectUDP(HostName, Port) != true)
		return false;

	return true;
}


// Disconnect from MotionVenus
bool UFOHEART_C1BPLibrary::FOHEART_C1Disconnect(AFOHEARTController *Controller)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	Controller->Disconnect();

	return true;
}

// Play BVH file
bool UFOHEART_C1BPLibrary::FOHEART_C1BVHPlay(AFOHEARTController *Controller, FString BVHFileName, bool bEndless, bool bReference, bool bDisplacement, EFOHEARTMotionLineFormatEnum MotionLineFormat)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	Controller->bReference = bReference;
	Controller->bDisplacement = bDisplacement;
	if (MotionLineFormat == EFOHEARTMotionLineFormatEnum::VE_FOHEART_C1)
		Controller->MotionLineFormat = FOHEART_C1;
	else
		Controller->MotionLineFormat = Standard;

	if (Controller->Play(BVHFileName, bEndless) != true)
		return false;

	return true;
}

// Pause playing BVH file
bool UFOHEART_C1BPLibrary::FOHEART_C1BVHPause(AFOHEARTController *Controller, bool bPause)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	return (Controller->Pause(bPause));
}

// Rewind BVH file player
bool UFOHEART_C1BPLibrary::FOHEART_C1BVHRewind(AFOHEARTController *Controller)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}

	return (Controller->Rewind());
}

//Read all quat data for bones.
void UFOHEART_C1BPLibrary::FOHEART_C1ReadBoneQuat(USkeletalMeshComponent *Mesh, FName TargetBoneName,
	FQuat& RefRelQuat, FQuat& RefGloQuat, FQuat& RefParentFloQuat)
{
	if (Mesh != nullptr && Mesh->SkeletalMesh != nullptr)
	{
		const FReferenceSkeleton& RefSkeleton(Mesh->SkeletalMesh->RefSkeleton);
		int32 RefBoneIndex = Mesh->GetBoneIndex(TargetBoneName);

		if (RefBoneIndex != INDEX_NONE)
		{
			RefRelQuat = RefSkeleton.GetRefBonePose()[RefBoneIndex].GetRotation();
			FRotator RefRotator = RefRelQuat.Rotator();
			//RefGloVector = RefSkeleton.GetRefBonePose()[RefBoneIndex].GetTranslation();
		}
		if (RefBoneIndex != INDEX_NONE)
		{
			int32 RefBoneIndexParent = RefSkeleton.GetParentIndex(RefBoneIndex);
			if (RefBoneIndexParent != INDEX_NONE)
			{
				while (RefBoneIndexParent != INDEX_NONE)
				{
					FQuat tmpQuat = RefSkeleton.GetRefBonePose()[RefBoneIndexParent].GetRotation();
					RefParentFloQuat = tmpQuat * RefParentFloQuat;
					RefBoneIndexParent = RefSkeleton.GetParentIndex(RefBoneIndexParent);
				}
			}
		}

		while (RefBoneIndex != INDEX_NONE)
		{
			FQuat tmpQuat = RefSkeleton.GetRefBonePose()[RefBoneIndex].GetRotation();
			RefGloQuat = tmpQuat * RefGloQuat;
			RefBoneIndex = RefSkeleton.GetParentIndex(RefBoneIndex);
		}
	}
}

// Read motion as array
bool UFOHEART_C1BPLibrary::FOHEART_C1ReadArray(AFOHEARTController *Controller, FString ActorName,
	USkeletalMeshComponent *Mesh, TArray<FVector> &Translation,
	TArray<FRotator> &Rotation, TArray<FQuat>& LeftHandQuat,
	TArray<FQuat>& RightHandQuat,
	TArray<FFOHEART_C1BoneMapStruct> BodyBoneMap,
	TArray<FFOHEART_C1HandBoneMapStruct> LeftHandBoneMap,
	TArray<FFOHEART_C1HandBoneMapStruct> RightHandBoneMap)
{
	if (Controller == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Controller is invalid.")));
		}
		return false;
	}
	else if ((Controller->bConnected == false) && (Controller->bPlay == false))
	{
		return false;
	}

	//ȡ
	float fData[PACKAGE_MAX_LEN];
	bool bRet = Controller->GetFloatBuffer(ActorName, fData);
	if (!bRet) return false;

	for (int32 BoneIndex = 0; BoneIndex < BodyBoneMap.Num(); BoneIndex++)
	{
		EFOHEART_C1BonesEnum BVHBoneEnum = BodyBoneMap[BoneIndex].FoheartBoneName;
		FName TargetBoneName = BodyBoneMap[BoneIndex].TargetBoneName;
		FVector TempTranslation;
		FRotator TempRotation;


		FQuat RefRelQuat(FQuat::Identity);
		FQuat RefGloQuat(FQuat::Identity);	//ǰȫԪ
		FQuat RefParentFloQuat(FQuat::Identity);//ǰĸڵȫԪ
		FOHEART_C1ReadBoneQuat(Mesh, TargetBoneName, RefRelQuat, RefGloQuat, RefParentFloQuat);


		int32 BVHBoneIndex = (int32)BVHBoneEnum;
		if (BVHBoneIndex >= Controller->Skeleton.BoneNr)
		{
			if (GEngine)
			{
				GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Boneindex %d exceeds maximum available bones %d."), BVHBoneIndex, Controller->Skeleton.BoneNr));
			}
			TempRotation.Yaw = TempRotation.Pitch = TempRotation.Roll = 0;
			TempTranslation.X = TempTranslation.Y = TempTranslation.Z = 0;
			continue;
		}
		int32 FloatsPerBone = 6; // 3 for x,y,z translation and 3 for x,y,z rotation
		if (Controller->bDisplacement == false)
		{
			FloatsPerBone = 3;	// If there is no displacement (translation) info we have only 3 floats for rotation left
		}
		if ((BVHBoneIndex * FloatsPerBone) > Controller->FloatCount)
		{
			TempRotation.Yaw = TempRotation.Pitch = TempRotation.Roll = 0;
			TempTranslation.X = TempTranslation.Y = TempTranslation.Z = 0;
			continue;
		}

		//任
		float XR = 0.0f, YR = 0.0f, ZR = 0.0f;
		float XP = 0.0f, YP = 0.0f, ZP = 0.0f;

		{
			//float *f = Controller->MotionLineMap[ActorName];
			XR = fData[(BVHBoneIndex * FloatsPerBone) + Controller->Skeleton.Bones[BVHBoneIndex].XRot];
			YR = fData[(BVHBoneIndex * FloatsPerBone) + Controller->Skeleton.Bones[BVHBoneIndex].YRot];
			ZR = fData[(BVHBoneIndex * FloatsPerBone) + Controller->Skeleton.Bones[BVHBoneIndex].ZRot];

			XP = fData[(BVHBoneIndex * FloatsPerBone) + Controller->Skeleton.Bones[BVHBoneIndex].XPos];
			YP = fData[(BVHBoneIndex * FloatsPerBone) + Controller->Skeleton.Bones[BVHBoneIndex].YPos];
			ZP = fData[(BVHBoneIndex * FloatsPerBone) + Controller->Skeleton.Bones[BVHBoneIndex].ZPos];
		}

		// Calculate Rotation Matrix and map to Quaternion
		FQuat BVHQuat = CalculateQuat(XR, YR, ZR, Controller->Skeleton.Bones[BVHBoneIndex].RotOrder);
		FQuat BVHVector(XP, YP, ZP, 0);

		//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString("--------------++++++-----------------"));
		BVHVector.Y = -BVHVector.Y;
		//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString("--ori:" + BVHVector.ToString()));
		BVHVector = RefParentFloQuat.Inverse() * BVHVector * RefParentFloQuat;

		//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString("--cal:" + BVHVector.ToString()));
		TempTranslation = FVector(BVHVector.X*100.0f, BVHVector.Y*100.0f, BVHVector.Z*100.0f);
		//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString("--------------++++++-----------------"));



		if (BVHBoneIndex == (int32)EFOHEART_C1BonesEnum::VE_LeftHand)
			UE_LOG(LogTemp, Log, TEXT("----1---- %f %f %f %f"), BVHQuat.X, BVHQuat.Y, BVHQuat.Z, BVHQuat.W);

		BVHQuat.Y = -BVHQuat.Y;
		BVHQuat = RefGloQuat.Inverse() * BVHQuat *RefGloQuat;

		if (BVHBoneIndex == (int32)EFOHEART_C1BonesEnum::VE_LeftHand)
			UE_LOG(LogTemp, Log, TEXT("----2---- %f %f %f %f"), RefGloQuat.X, RefGloQuat.Y, RefGloQuat.Z, RefGloQuat.W);

		FQuat Quat = RefRelQuat * BVHQuat;

		if (BVHBoneIndex == (int32)EFOHEART_C1BonesEnum::VE_LeftHand)
			UE_LOG(LogTemp, Log, TEXT("----3---- %f %f %f %f"), BVHQuat.X, BVHQuat.Y, BVHQuat.Z, BVHQuat.W);


		// Convert to Rotator
		TempRotation = Quat.Rotator();
		TempRotation.Normalize();


		Translation.Add(TempTranslation);
		Rotation.Add(TempRotation);
	}

	//

	for (int32 BoneIndex = 0; BoneIndex < LeftHandBoneMap.Num(); BoneIndex++)
	{
		EFOHEART_C1HandBonesEnum BVHBoneEnum = LeftHandBoneMap[BoneIndex].FoheartBoneName;
		FName TargetBoneName = LeftHandBoneMap[BoneIndex].TargetBoneName;
		FQuat tempQuat = FQuat::Identity;

		FQuat RefRelQuat(FQuat::Identity);
		FQuat RefGloQuat(FQuat::Identity);	//ǰȫԪ
		FQuat RefParentFloQuat(FQuat::Identity);//ǰĸڵȫԪ
		FOHEART_C1ReadBoneQuat(Mesh, TargetBoneName, RefRelQuat, RefGloQuat, RefParentFloQuat);


		int32 BVHBoneIndex = (int32)BVHBoneEnum;
		if (BVHBoneIndex >= Controller->Skeleton.BoneNr)
		{
			if (GEngine)
			{
				GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Boneindex %d exceeds maximum available bones %d."), BVHBoneIndex, Controller->Skeleton.BoneNr));
			}
			continue;
		}

		//任
		float XR = 0.0f, YR = 0.0f, ZR = 0.0f, WR = 1.0f;
		{
			//float *f = Controller->MotionLineMap[ActorName];
			XR = fData[(23 * 6 + BVHBoneIndex * 4) + 0];
			YR = fData[(23 * 6 + BVHBoneIndex * 4) + 1];
			ZR = fData[(23 * 6 + BVHBoneIndex * 4) + 2];
			WR = fData[(23 * 6 + BVHBoneIndex * 4) + 3];
		}

		// Calculate Rotation Matrix and map to Quaternion
		FQuat BVHQuat(XR, YR, ZR, WR);

		if (BVHBoneIndex == (int32)EFOHEART_C1HandBonesEnum::VS_Index3)
			UE_LOG(LogTemp, Log, TEXT("----11---- %f %f %f %f"), BVHQuat.X, BVHQuat.Y, BVHQuat.Z, BVHQuat.W);

		BVHQuat.X = -BVHQuat.X;
		BVHQuat.Y = BVHQuat.Y;
		BVHQuat.Z = -BVHQuat.Z;
		BVHQuat = RefGloQuat.Inverse() * BVHQuat *RefGloQuat;
		if (BVHBoneIndex == (int32)EFOHEART_C1HandBonesEnum::VS_Index3)
			UE_LOG(LogTemp, Log, TEXT("----22---- %f %f %f %f"), RefGloQuat.X, RefGloQuat.Y, RefGloQuat.Z, RefGloQuat.W);

		//ChangeAxisMatchMap(BVHQuat, BVHQuat, MatchAxis_X.FOHEARTBoneAxis, MatchAxis_Y.FOHEARTBoneAxis, MatchAxis_Z.FOHEARTBoneAxis);
		FQuat Quat = RefRelQuat * BVHQuat;// *RefRelQuat.Inverse();

		if (BVHBoneIndex == (int32)EFOHEART_C1HandBonesEnum::VS_Index3)
			UE_LOG(LogTemp, Log, TEXT("----33---- %f %f %f %f"), BVHQuat.X, BVHQuat.Y, BVHQuat.Z, BVHQuat.W);

		LeftHandQuat.Add(Quat);
	}
	//

	for (int32 BoneIndex = 0; BoneIndex < RightHandBoneMap.Num(); BoneIndex++)
	{
		EFOHEART_C1HandBonesEnum BVHBoneEnum = RightHandBoneMap[BoneIndex].FoheartBoneName;
		FName TargetBoneName = RightHandBoneMap[BoneIndex].TargetBoneName;
		FQuat tempQuat = FQuat::Identity;

		FQuat RefRelQuat(FQuat::Identity);
		FQuat RefGloQuat(FQuat::Identity);	//ǰȫԪ
		FQuat RefParentFloQuat(FQuat::Identity);//ǰĸڵȫԪ
		FOHEART_C1ReadBoneQuat(Mesh, TargetBoneName, RefRelQuat, RefGloQuat, RefParentFloQuat);

		int32 BVHBoneIndex = (int32)BVHBoneEnum;
		if (BVHBoneIndex >= Controller->Skeleton.BoneNr)
		{
			if (GEngine)
			{
				GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Boneindex %d exceeds maximum available bones %d."), BVHBoneIndex, Controller->Skeleton.BoneNr));
			}
			continue;
		}

		//任
		float XR = 0.0f, YR = 0.0f, ZR = 0.0f, WR = 1.0f;
		{
			//float *f = Controller->MotionLineMap[ActorName];
			XR = fData[(23 * 6 + (int32)EFOHEART_C1HandBonesEnum::VS_Max * 4 + BVHBoneIndex * 4) + 0];
			YR = fData[(23 * 6 + (int32)EFOHEART_C1HandBonesEnum::VS_Max * 4 + BVHBoneIndex * 4) + 1];
			ZR = fData[(23 * 6 + (int32)EFOHEART_C1HandBonesEnum::VS_Max * 4 + BVHBoneIndex * 4) + 2];
			WR = fData[(23 * 6 + (int32)EFOHEART_C1HandBonesEnum::VS_Max * 4 + BVHBoneIndex * 4) + 3];
		}

		// Calculate Rotation Matrix and map to Quaternion
		FQuat BVHQuat(XR, YR, ZR, WR);


		BVHQuat.X = -BVHQuat.X;
		BVHQuat.Y = BVHQuat.Y;
		BVHQuat.Z = -BVHQuat.Z;
		BVHQuat = RefGloQuat.Inverse() * BVHQuat *RefGloQuat;

		//ChangeAxisMatchMap(BVHQuat, BVHQuat, MatchAxis_X.FOHEARTBoneAxis, MatchAxis_Y.FOHEARTBoneAxis, MatchAxis_Z.FOHEARTBoneAxis);
		FQuat Quat = RefRelQuat * BVHQuat;// *RefRelQuat.Inverse();

		RightHandQuat.Add(Quat);
	}


	//FHitResult r;
	//Mesh->K2_SetWorldLocation(Translation[0], false, r, false);

	return true;
}

// Get local bone rotation from mesh
bool UFOHEART_C1BPLibrary::FOHEART_C1GetLocalBoneRotation(USkeletalMeshComponent *Mesh, FRotator& Rotation, int32 BoneIndex)
{
	if (Mesh == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Mesh is invalid.")));
		}
		Rotation.Yaw = Rotation.Pitch = Rotation.Roll = 0;
		return false;
	}

#if (ENGINE_MAJOR_VERSION >= 4) && (ENGINE_MINOR_VERSION >= 13)
	int32 BoneIndexMax = Mesh->BoneSpaceTransforms.Num();
#else
	int32 BoneIndexMax = Mesh->LocalAtoms.Num();
#endif
	if (BoneIndex > BoneIndexMax)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("BoneIndex %d exceeds maximum available bones %d."), BoneIndex, BoneIndexMax));
		}
		Rotation.Yaw = Rotation.Pitch = Rotation.Roll = 0;
		return false;
	}

#if (ENGINE_MAJOR_VERSION >= 4) && (ENGINE_MINOR_VERSION >= 13)
	Rotation = Mesh->BoneSpaceTransforms[BoneIndex].Rotator();
#else
	Rotation = Mesh->LocalAtoms[BoneIndex].Rotator();
#endif

	return true;
}

// Get local bone location from mesh
bool UFOHEART_C1BPLibrary::FOHEART_C1GetLocalBoneLocation(USkeletalMeshComponent *Mesh, FVector& Location, int32 BoneIndex)
{
	if (Mesh == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Mesh is invalid.")));
		}
		Location.X = Location.Y = Location.Z = 0;
		return false;
	}

#if (ENGINE_MAJOR_VERSION >= 4) && (ENGINE_MINOR_VERSION >= 13)
	int32 BoneIndexMax = Mesh->BoneSpaceTransforms.Num();
#else
	int32 BoneIndexMax = Mesh->LocalAtoms.Num();
#endif
	if (BoneIndex > BoneIndexMax)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("BoneIndex %d exceeds maximum available bones %d."), BoneIndex, BoneIndexMax));
		}
		Location.X = Location.Y = Location.Z = 0;
		return false;
	}

#if (ENGINE_MAJOR_VERSION >= 4) && (ENGINE_MINOR_VERSION >= 13)
	Location = Mesh->BoneSpaceTransforms[BoneIndex].GetLocation();
#else
	Location = Mesh->LocalAtoms[BoneIndex].GetLocation();
#endif
	return true;
}

bool UFOHEART_C1BPLibrary::FOHEART_C1GetReferencePoseLocalBoneRotation(USkeletalMeshComponent *Mesh, FRotator& Rotation, int32 BoneIndex)
{
	if (Mesh == nullptr && Mesh->SkeletalMesh == nullptr)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Mesh is invalid.")));
		}
		Rotation.Yaw = Rotation.Pitch = Rotation.Roll = 0;
		return false;
	}

#if (ENGINE_MAJOR_VERSION >= 4) && (ENGINE_MINOR_VERSION >= 13)
	int32 BoneIndexMax = Mesh->BoneSpaceTransforms.Num();
#else
	int32 BoneIndexMax = Mesh->LocalAtoms.Num();
#endif
	if (BoneIndex > BoneIndexMax)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("BoneIndex %d exceeds maximum available bones %d."), BoneIndex, BoneIndexMax));
		}
		Rotation.Yaw = Rotation.Pitch = Rotation.Roll = 0;
		return false;
	}

	const FReferenceSkeleton& refskel(Mesh->SkeletalMesh->RefSkeleton);
	FQuat Quat = refskel.GetRefBonePose()[BoneIndex].GetRotation();
	Rotation = Quat.Rotator();

	return true;
}

// Negate Yaw, Pitch and Roll in rotation vector
FRotator UFOHEART_C1BPLibrary::FOHEART_C1NegateRotation(FRotator Rotation)
{
	FRotator NewRotation;
	NewRotation.Yaw = -Rotation.Yaw;
	NewRotation.Pitch = -Rotation.Pitch;
	NewRotation.Roll = -Rotation.Roll;

	return NewRotation;
}

// Return BVH Bone Index
int32 UFOHEART_C1BPLibrary::FOHEART_C1GetBVHBoneIndex(EFOHEART_C1BonesEnum FoheartBoneName)
{
	return (int32)FoheartBoneName;
}

void UFOHEART_C1BPLibrary::ChangeAxisMatchMap(FQuat qOri, FQuat& qNew, EFOHEARTBoneAxisEnum matchX, EFOHEARTBoneAxisEnum matchY, EFOHEARTBoneAxisEnum matchZ)
{
	switch (matchX)
	{
	case EFOHEARTBoneAxisEnum::VE_Axis_X:
		qNew.X = qOri.X;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_Y:
		qNew.X = qOri.Y;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_Z:
		qNew.X = qOri.Z;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iX:
		qNew.X = -qOri.X;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iY:
		qNew.X = -qOri.Y;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iZ:
		qNew.X = -qOri.Z;
		break;
	}

	switch (matchY)
	{
	case EFOHEARTBoneAxisEnum::VE_Axis_X:
		qNew.Y = qOri.X;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_Y:
		qNew.Y = qOri.Y;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_Z:
		qNew.Y = qOri.Z;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iX:
		qNew.Y = -qOri.X;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iY:
		qNew.Y = -qOri.Y;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iZ:
		qNew.Y = -qOri.Z;
		break;
	}
	switch (matchZ)
	{
	case EFOHEARTBoneAxisEnum::VE_Axis_X:
		qNew.Z = qOri.X;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_Y:
		qNew.Z = qOri.Y;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_Z:
		qNew.Z = qOri.Z;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iX:
		qNew.Z = -qOri.X;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iY:
		qNew.Z = -qOri.Y;
		break;
	case EFOHEARTBoneAxisEnum::VE_Axis_iZ:
		qNew.Z = -qOri.Z;
		break;
	}

}
