// Copyright 1998-2017 Epic Games, Inc. All Rights Reserved.

#include "FOHEARTController.h"
#include "FOHEART.h"

// Sets default values
AFOHEARTController::AFOHEARTController()
{
	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;
	frameNum = 0;
}

// Called when the game starts or when spawned
void AFOHEARTController::BeginPlay()
{
	Super::BeginPlay();
}

// Called every frame
void AFOHEARTController::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	// Read each tick a new motionline. Additionally motionlines are just discarded.
	if ((bConnected == true) || (bPlay == true))
	{
		uint32 BytesPending = 0;

		if (bConnected == true)
			if (ReceiverSocket->HasPendingData(BytesPending) != true)
				BytesPending = 0;

		if ((BytesPending) || (bPlay == true))
		{
#define SIZEOFDATA 1500
			uint8 Data[SIZEOFDATA];
			int32 BytesRead = 0;
			bool bMotionLineBeginFound = false;
			bool bMotionLineEndFound = false;

			if (bConnected == true) // Network connection?
			{
				if (ReceiverSocket->Recv(Data, SIZEOFDATA, BytesRead) != true)
					BytesRead = 0;
			}
			else if (bPlay == true)	// Player connected ?
			{
				DeltaTimeAdded += DeltaTime;
				if (FrameTime < DeltaTimeAdded) // Synchronize to FrameTime
				{
					DeltaTimeAdded -= FrameTime;
					while (DeltaTimeAdded > FrameTime) // We are too slow => Skip frames
					{
						DeltaTimeAdded -= FrameTime;
						MotionLinePointer++;
						if ((MotionLinePointer >= (MotionLineOffset + Frames)) && (bEndless == true))
							MotionLinePointer = MotionLineOffset;
					}

					if (MotionLinePointer < (MotionLineOffset + Frames))
					{
						int32 CurrentMotionLine = MotionLinePointer;
						if (SIZEOFDATA >= PlayerMotionLines[CurrentMotionLine].Len() + 1) // Sanity check
						{
							// Convert string based motion line back to an array of chars to pump the line through the parser (Little bit ugly...)
							memcpy(Data, TCHAR_TO_ANSI(*PlayerMotionLines[CurrentMotionLine]), PlayerMotionLines[CurrentMotionLine].Len());
							Data[PlayerMotionLines[CurrentMotionLine].Len()] = '\n';
							BytesRead = PlayerMotionLines[CurrentMotionLine].Len() + 1;
						}
						MotionLinePointer++; // Increment line for next time;
						if ((MotionLinePointer >= (MotionLineOffset + Frames)) && (bEndless == true))
							MotionLinePointer = MotionLineOffset;
					}
				}
			}

			if (BytesRead)
			{
				if (MotionLineFormat == FOHEART_C1)
				{
					int dataIndex = 0;

					uint16_t motionVenusProtoVer = 0;
					memcpy(&motionVenusProtoVer, Data + dataIndex, sizeof(motionVenusProtoVer));
					dataIndex += sizeof(motionVenusProtoVer);
					//if (GEngine) GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("ProtoVer:%d"), motionVenusProtoVer));

					uint8_t nameLength = Data[dataIndex];
					dataIndex += sizeof(nameLength);

					char *newName = new char[nameLength + 1];
					memcpy(newName, Data + dataIndex, nameLength);
					newName[nameLength] = '\0';
					FString  mActorName = newName;
					delete newName;
					dataIndex += nameLength;

					uint32_t suitNumber = 0;
					memcpy(&suitNumber, Data + dataIndex, sizeof(suitNumber));
					dataIndex += sizeof(suitNumber);
					//if (GEngine) GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("SuitNum:0x%x"), suitNumber));

					uint8_t    suitType = Data[dataIndex];
					dataIndex += sizeof(suitType);
					//if (GEngine) GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("SuitType:%d"), suitType));

					uint32_t   frameNumber = 0;
					memcpy(&frameNumber, Data + dataIndex, sizeof(frameNumber));
					dataIndex += sizeof(frameNumber);
					//if (GEngine) GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("frameNum:%d"), frameNumber));
					//UE_LOG(LogInit, Log, TEXT("Actor: %s %d"),*mActorName, frameNumber);
					uint8_t boneCount = Data[dataIndex];
					dataIndex += sizeof(boneCount);
					//if (GEngine) GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("boneCount:%d"), boneCount));

					//׽ڵ
					uint8_t lefthandNodeCount = Data[dataIndex];
					dataIndex += sizeof(lefthandNodeCount);
					//UE_LOG(LogInit, Log, TEXT("Left: %u"),lefthandNodeCount);

					//׽ڵ
					uint8_t righthandNodeCount = Data[dataIndex];
					dataIndex += sizeof(righthandNodeCount);
					//UE_LOG(LogInit, Log, TEXT("Right: %u"), righthandNodeCount);
					frameNum++;

					if (frameNum % 60 == 0)
					{
						//UE_LOG(LogInit, Log, TEXT("freq:%d"), frameNumber - lastSendFrameNum);
						lastSendFrameNum = frameNumber;
					}

					dataIndex = 128;

					float *f;
					if (!MotionLineMap.Contains(mActorName))
					{
						float* &refNewMotionLine = MotionLineMap.FindOrAdd(mActorName);
						refNewMotionLine = new float[PACKAGE_MAX_LEN]{ 0x00 };
						f = MotionLineMap[mActorName];
					}
					else
					{
						f = MotionLineMap[mActorName];
					}
					FScopeLock ScopeLock(&Mutex);
					uint16_t index = 0;
					for (uint8_t i = 0; i < boneCount; i++)
					{
						int32_t pos[3] = { 0 };
						memcpy(pos, Data + dataIndex, sizeof(pos));

						const uint32_t LocScale = (1 << 16);
						index = i * 6 + 0;
						f[index] = (((float)pos[0]) / LocScale);
						index = i * 6 + 1;
						f[index] = (((float)pos[1]) / LocScale);
						index = i * 6 + 2;
						f[index] = (((float)pos[2]) / LocScale);

						/*	MotionLine[i * 6 + 0] = (float)pos[0] / (1 << 16) * 100;
						MotionLine[i * 6 + 1] = (float)pos[1] / (1 << 16) * 100;
						MotionLine[i * 6 + 2] = (float)pos[2] / (1 << 16) * 100;*/
						dataIndex += sizeof(pos);

						int16_t euler[3] = { 0 };
						memcpy(euler, Data + dataIndex, sizeof(euler));

						const uint16_t EulerScale = (1 << 7);
						index = i * 6 + 3;
						f[index] = (((float)euler[0]) / EulerScale);
						index = i * 6 + 4;
						f[index] = (((float)euler[1]) / EulerScale);
						index = i * 6 + 5;
						f[index] = (((float)euler[2]) / EulerScale);

						/*	MotionLine[i * 6 + 3] = (float)euler[0] / (1 << 7);
						MotionLine[i * 6 + 4] = (float)euler[1] / (1 << 7);
						MotionLine[i * 6 + 5] = (float)euler[2] / (1 << 7);*/
						dataIndex += sizeof(euler);

						if (GEngine && i == 21)
						{
								/*GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("%d pos:%f %f %f euler:%f %f %f"),
									i, f[i * 6 + 0], f[i * 6 + 1], f[i * 6 + 2],
									f[i * 6 + 3], f[i * 6 + 4], f[i * 6 + 5]
								));*/
						}
					}

					const uint32_t QuatScale = (1 << 13);
					////UE_LOG(LogInit, Log, TEXT("---1--: %u"), dataIndex);
					//UE_LOG(LogInit, Log, TEXT("---111--: %u"), index);
					//
					for (int i = 0; i < lefthandNodeCount; i++)
					{
						int16_t quat[4] = { 0 };
						memcpy(quat, Data + dataIndex, sizeof(quat));

						index += 1;
						f[index] = (((float)quat[0]) / QuatScale);
						index += 1;
						f[index] = (((float)quat[1]) / QuatScale);
						index += 1;
						f[index] = (((float)quat[2]) / QuatScale);
						index += 1;
						f[index] = (((float)quat[3]) / QuatScale);

						dataIndex += sizeof(quat);
					}
					//UE_LOG(LogInit, Log, TEXT("---222--: %u"), index);
					//UE_LOG(LogInit, Log, TEXT("---2--: %u"), dataIndex);

					//
					for (int i = 0; i < righthandNodeCount; i++)
					{
						int16_t quat[4] = { 0 };
						memcpy(quat, Data + dataIndex, sizeof(quat));

						index += 1;
						f[index] = (((float)quat[0]) / QuatScale);
						index += 1;
						f[index] = (((float)quat[1]) / QuatScale);
						index += 1;
						f[index] = (((float)quat[2]) / QuatScale);
						index += 1;
						f[index] = (((float)quat[3]) / QuatScale);

						dataIndex += sizeof(quat);
						//UE_LOG(LogInit, Log, TEXT("-------: %u"), index);
					}
				}
				
				if (bConnected)
				{
					// If we get more data then we can process per tick, read socket till no more data is left and ignore it
					//while (ReceiverSocket->HasPendingData(BytesPending))
					//{
					//	static uint32 cc = 0;
					//	ReceiverSocket->Recv(Data, SIZEOFDATA, BytesRead);
					//	UE_LOG(LogInit, Log, TEXT("Data skipped: %u %u"), BytesRead, cc++);
					//}
				}
			} // End Recv
		} // End HasPendingData
	} // End bConnected
}

// Connect to BVH server ( MotionVenus )
bool AFOHEARTController::Connect(FString HostName, int32 Port)
{
	if (bConnected == true)
		return false;

	// Create TCP Socket and connect
	ISocketSubsystem* const SocketSubSystem = ISocketSubsystem::Get();
	if (SocketSubSystem)
	{
		ReceiverSocket = SocketSubSystem->CreateSocket(NAME_Stream, TEXT("MotionVenusSocket"), false);
		if (ReceiverSocket == NULL)
			return false;

		// Is HostName already a valid IPv4 address?		
		bool bValidAddress = false;
		uint32 IP = 0;
		FIPv4Address IPv4Address;
		if (true == FIPv4Address::Parse(HostName, IPv4Address))
		{
			IP = IPv4Address.Value;
			bValidAddress = true;
		}
		else
		{	// Resolve HostName
			auto ResolveInfo = SocketSubSystem->GetHostByName(TCHAR_TO_ANSI(*HostName));
			while (!ResolveInfo->IsComplete());

			if (ResolveInfo->GetErrorCode() == 0)
			{
				const FInternetAddr* Addr = &ResolveInfo->GetResolvedAddress();
				Addr->GetIp(IP);
				bValidAddress = true;
			}
		}

		// Connect
		if (bValidAddress)
		{
			TSharedRef<FInternetAddr> InetAddr = SocketSubSystem->CreateInternetAddr();
			InetAddr->SetIp(IP);
			InetAddr->SetPort(Port);
			bConnected = ReceiverSocket->Connect(*InetAddr);
		}
	}

	return bConnected;
}

// Connect to BVH server ( MotionVenus )
bool AFOHEARTController::ConnectUDP(FString HostName, int32 Port)
{
	if (bConnected == true)
		return false;

	// Create TCP Socket and connect
	ISocketSubsystem* const SocketSubSystem = ISocketSubsystem::Get();
	if (SocketSubSystem)
	{
		ReceiverSocket = SocketSubSystem->CreateSocket(NAME_DGram, TEXT("MotionVenusSocket"), false);
		if (ReceiverSocket == NULL)
			return false;

		// Is HostName already a valid IPv4 address?		
		bool bValidAddress = false;
		uint32 IP = 0;
		FIPv4Address IPv4Address;
		if (true == FIPv4Address::Parse(HostName, IPv4Address))
		{
			IP = IPv4Address.Value;
			bValidAddress = true;
		}
		else
		{	// Resolve HostName
			auto ResolveInfo = SocketSubSystem->GetHostByName(TCHAR_TO_ANSI(*HostName));
			while (!ResolveInfo->IsComplete());

			if (ResolveInfo->GetErrorCode() == 0)
			{
				const FInternetAddr* Addr = &ResolveInfo->GetResolvedAddress();
				Addr->GetIp(IP);
				bValidAddress = true;
			}
		}

		// Connect
		if (bValidAddress)
		{
			/*TSharedRef<FInternetAddr> InetAddr = SocketSubSystem->CreateInternetAddr();
			InetAddr->SetIp(IP);
			InetAddr->SetPort(Port);
			bConnected = ReceiverSocket->Connect(*InetAddr);*/

			FUdpSocketBuilder builder = FUdpSocketBuilder(TEXT("MCUDPSocket"))
				.BoundToPort(Port).BoundToAddress(IPv4Address)
				.AsNonBlocking().AsReusable()
				.WithReceiveBufferSize(1024 + 512);
			ReceiverSocket = builder.Build();
			bConnected = true;
		}
	}

	return bConnected;
}

// Called when the game stops
void AFOHEARTController::BeginDestroy()
{
	Super::BeginDestroy();

	Disconnect();
}

// Disconnect from BVH server
void AFOHEARTController::Disconnect(void)
{
	if (bConnected == true)
	{
		bConnected = false;
		ReceiverSocket->Close();
	}
}

// BVH Player
bool AFOHEARTController::Play(FString BVHFileName, bool bPlayEndless)
{
	// Set to default
	bPlayerInitialized = false;
	bPlay = false;
	PlayerMotionLines.Empty();

	// Load BVH file	
	if (FFileHelper::LoadANSITextFileToStrings(*(FPaths::GameDir() + FString("Content/") + BVHFileName), NULL, PlayerMotionLines) != true)
	{
		if (GEngine)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Could not load BVH file:%s"), *(FPaths::GameDir() + FString("Content/") + BVHFileName)));
		}
		return false;
	}

	// Parse BVH file, get frame time and beginning of motion line part
	for (int32 i = 0; i < PlayerMotionLines.Num(); i++)
	{
		PlayerMotionLines[i].Trim();

		if (PlayerMotionLines[i].StartsWith(TEXT("Frames:"), ESearchCase::IgnoreCase) == true)
		{
			TArray<FString> Words;
			if (PlayerMotionLines[i].ParseIntoArray(Words, TEXT(" "), false) >= 2)
			{
				Frames = FCString::Atoi(*Words[1]);
			}
		}
		else if (PlayerMotionLines[i].StartsWith(TEXT("Frame Time:"), ESearchCase::IgnoreCase) == true)
		{
			TArray<FString> Words;
			if (PlayerMotionLines[i].ParseIntoArray(Words, TEXT(" "), false) >= 2)
			{
				FrameTime = FCString::Atof(*Words[2]);
				if (FrameTime == 0)
					FrameTime = 0.010f;  // Default to a common value
			}
			if ((i + 1) <= PlayerMotionLines.Num())
			{
				MotionLineOffset = i + 1;
				MotionLinePointer = MotionLineOffset;
				bPlayerInitialized = true;
				bPlay = true;
				AFOHEARTController::bEndless = bPlayEndless;
				if (Frames > PlayerMotionLines.Num() - (i + 1))
					Frames = PlayerMotionLines.Num() - (i + 1);
				DeltaTimeAdded = FrameTime; // Play motion immediately with next frame
			}
		}
	}
	return bPlayerInitialized;
}

bool AFOHEARTController::Pause(bool bPause)
{
	if (bPlayerInitialized == true)
		bPlay = bPause;
	return bPlay;
}

bool AFOHEARTController::Rewind()
{
	if (bPlayerInitialized == true)
		MotionLinePointer = MotionLineOffset;
	return bPlayerInitialized;
}

bool AFOHEARTController::GetFloatBuffer(FString ActorName, float * pFloatData)
{
	float ** ppData = MotionLineMap.Find(ActorName);
	if (ppData)
	{
		FScopeLock ScopeLock(&Mutex);
		memcpy(pFloatData, *ppData, sizeof(float)*PACKAGE_MAX_LEN);
		return true;
	}
	else
		return false;

}
